@@ -4,6 +4,7 @@
 
 #include "vtkDataArray.h"
 #include "vtkEndian.h"
+#include "vtkErrorCode.h"
 #include "vtkImageData.h"
 #include "vtkObjectFactory.h"
 #include "vtkPointData.h"
@@ -75,7 +76,7 @@ void PNGReadCallback(png_structp pngPtr, png_bytep output, png_size_t length)
   // Advance cursor
   input->position += length;
 }
-};
+}
 
 class vtkPNGReader::vtkInternals
 {
@@ -84,6 +85,14 @@ public:
   typedef std::vector<std::pair<std::string, std::string>>::iterator TextKeyValueIterator;
   vtkNew<vtkStringArray> TextKeys;
   vtkNew<vtkStringArray> TextValues;
+
+  vtkPNGReader* const PNGReader = nullptr;
+
+  vtkInternals(vtkPNGReader* reader)
+    : PNGReader{ reader }
+  {
+  }
+
   void ReadTextChunks(png_structp png_ptr, png_infop info_ptr)
   {
     png_textp text_ptr;
@@ -121,7 +130,7 @@ public:
     bool is_png = !png_sig_cmp(header, 0, 8);
     if (!is_png)
     {
-      vtkErrorWithObjectMacro(nullptr, << "Unknown file type! Not a PNG file!");
+      vtkErrorWithObjectMacro(this->PNGReader, << "Unknown file type! Not a PNG file!");
     }
     return is_png;
   }
@@ -132,7 +141,7 @@ public:
     unsigned char header[8];
     if (fread(header, 1, 8, fp) != 8)
     {
-      vtkErrorWithObjectMacro(nullptr,
+      vtkErrorWithObjectMacro(this->PNGReader,
         "PNGReader error reading file."
           << " Premature EOF while reading header.");
       return false;
@@ -146,7 +155,8 @@ public:
     unsigned char header[8];
     if (length < 8)
     {
-      vtkErrorWithObjectMacro(nullptr, "MemoryBuffer is too short, could not read the header");
+      vtkErrorWithObjectMacro(
+        this->PNGReader, "MemoryBuffer is too short, could not read the header");
       return false;
     }
     std::copy(buffer, buffer + 8, header);
@@ -158,21 +168,21 @@ public:
     pngPtr = png_create_read_struct(PNG_LIBPNG_VER_STRING, (png_voidp) nullptr, nullptr, nullptr);
     if (!pngPtr)
     {
-      vtkErrorWithObjectMacro(nullptr, "Out of memory.");
+      vtkErrorWithObjectMacro(this->PNGReader, "Out of memory.");
       return false;
     }
     infoPtr = png_create_info_struct(pngPtr);
     if (!infoPtr)
     {
       png_destroy_read_struct(&pngPtr, (png_infopp) nullptr, (png_infopp) nullptr);
-      vtkErrorWithObjectMacro(nullptr, "Out of memory.");
+      vtkErrorWithObjectMacro(this->PNGReader, "Out of memory.");
       return false;
     }
     endInfo = png_create_info_struct(pngPtr);
     if (!endInfo)
     {
       png_destroy_read_struct(&pngPtr, &infoPtr, (png_infopp) nullptr);
-      vtkErrorWithObjectMacro(nullptr, "Unable to read PNG file!");
+      vtkErrorWithObjectMacro(this->PNGReader, "Unable to read PNG file!");
       return false;
     }
     return true;
@@ -215,7 +225,7 @@ public:
 //------------------------------------------------------------------------------
 vtkPNGReader::vtkPNGReader()
 {
-  this->Internals = new vtkInternals();
+  this->Internals = new vtkInternals(this);
   this->ReadSpacingFromFile = false;
 }
 
@@ -239,6 +249,7 @@ void vtkPNGReader::ExecuteInformation()
     if (!impl->CheckBufferHeader(memBuffer, this->GetMemoryBufferLength()))
     {
       vtkErrorMacro("Invalid MemoryBuffer header: not a PNG file");
+      this->SetErrorCode(vtkErrorCode::UnrecognizedFileTypeError);
       return;
     }
   }
@@ -249,18 +260,21 @@ void vtkPNGReader::ExecuteInformation()
     if (this->InternalFileName == nullptr)
     {
       vtkErrorMacro("A filename must be specified");
+      this->SetErrorCode(vtkErrorCode::NoFileNameError);
       return;
     }
     fp = vtksys::SystemTools::Fopen(this->InternalFileName, "rb");
     if (!fp)
     {
       vtkErrorMacro("Unable to open file " << this->InternalFileName);
+      this->SetErrorCode(vtkErrorCode::CannotOpenFileError);
       return;
     }
     if (!impl->CheckFileHeader(fp))
     {
       vtkErrorMacro("Invalid file header: not a PNG file");
       fclose(fp);
+      this->SetErrorCode(vtkErrorCode::FileFormatError);
       return;
     }
   }
@@ -369,6 +383,7 @@ void vtkPNGReader::vtkPNGReaderUpdate2(OT* outPtr, int* outExt, vtkIdType* outIn
     if (!impl->CheckBufferHeader(memBuffer, this->GetMemoryBufferLength()))
     {
       vtkErrorMacro("Invalid MemoryBuffer header: not a PNG file");
+      this->SetErrorCode(vtkErrorCode::FileFormatError);
       return;
     }
   }
@@ -379,12 +394,14 @@ void vtkPNGReader::vtkPNGReaderUpdate2(OT* outPtr, int* outExt, vtkIdType* outIn
     if (!fp)
     {
       vtkErrorMacro("Unable to open file " << this->InternalFileName);
+      this->SetErrorCode(vtkErrorCode::CannotOpenFileError);
       return;
     }
     if (!impl->CheckFileHeader(fp))
     {
       vtkErrorMacro("Invalid file header: not a PNG file");
       fclose(fp);
+      this->SetErrorCode(vtkErrorCode::FileFormatError);
       return;
     }
   }
@@ -515,6 +532,7 @@ void vtkPNGReader::ExecuteDataWithInformation(vtkDataObject* output, vtkInformat
   if (!this->GetMemoryBuffer() && this->InternalFileName == nullptr)
   {
     vtkErrorMacro(<< "Either a FileName, FilePrefix or MemoryBuffer must be specified.");
+    this->SetErrorCode(vtkErrorCode::NoFileNameError);
     return;
   }
 
@@ -532,6 +550,7 @@ void vtkPNGReader::ExecuteDataWithInformation(vtkDataObject* output, vtkInformat
     vtkTemplateMacro(this->vtkPNGReaderUpdate(data, (VTK_TT*)(outPtr)));
     default:
       vtkErrorMacro(<< "UpdateFromFile: Unknown data type");
+      this->SetErrorCode(vtkErrorCode::UnrecognizedFileTypeError);
   }
 }
 
