@@ -1,7 +1,7 @@
 // SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
 // SPDX-License-Identifier: BSD-3-Clause
 #include "vtkOpenGLRenderWindow.h"
-#include "vtk_glew.h"
+#include "vtk_glad.h"
 
 #include "vtkOpenGLHelper.h"
 
@@ -29,6 +29,7 @@
 #include "vtkPerlinNoise.h"
 #include "vtkRenderTimerLog.h"
 #include "vtkRendererCollection.h"
+#include "vtkRenderingOpenGLConfigure.h"
 #include "vtkShaderProgram.h"
 #include "vtkStringOutputWindow.h"
 #include "vtkTextureObject.h"
@@ -36,10 +37,27 @@
 #include "vtkTimerLog.h"
 #include "vtkUnsignedCharArray.h"
 
+#if defined(_WIN32)
+#include "vtkWin32OpenGLRenderWindow.h"
+#endif
+#if defined(VTK_USE_X)
+#include "vtkXOpenGLRenderWindow.h"
+#include "vtkglad/include/glad/glx.h"
+#endif
+#if defined(VTK_OPENGL_HAS_EGL)
+#include "vtkEGLRenderWindow.h"
+#include "vtkglad/include/glad/egl.h"
+#endif
+#include "vtkOSOpenGLRenderWindow.h"
+
+#include "vtksys/SystemTools.hxx"
+
 #include "BlueNoiseTexture64x64.h"
 #include "vtkTextureObjectVS.h" // a pass through shader
 
+#include <cstdlib>
 #include <sstream>
+#include <string>
 #include <type_traits>
 using std::ostringstream;
 
@@ -97,10 +115,10 @@ static const char* ResolveShader =
 
     for (int i = 0; i < samplecount; i++)
     {
-      vec4 sample = texelFetch(tex, itexcoords, i);
+      vec4 sampleValue = texelFetch(tex, itexcoords, i);
       // apply gamma correction and sum
-      accumulate += pow(sample.rgb, vec3(gamma));
-      alpha += sample.a;
+      accumulate += pow(sampleValue.rgb, vec3(gamma));
+      alpha += sampleValue.a;
     }
 
     // divide and reverse gamma correction
@@ -392,9 +410,11 @@ vtkOpenGLRenderWindow::vtkOpenGLRenderWindow()
   this->FramebufferFlipY = false;
 
   this->Initialized = false;
-  this->GlewInitValid = false;
 
-  this->MultiSamples = vtkOpenGLRenderWindowGlobalMaximumNumberOfMultiSamples;
+  this->MultiSamples = vtksys::SystemTools::HasEnv("VTK_TESTING")
+    ? 0
+    : vtkOpenGLRenderWindowGlobalMaximumNumberOfMultiSamples;
+
   delete[] this->WindowName;
   this->WindowName = new char[strlen(defaultWindowName) + 1];
   strcpy(this->WindowName, defaultWindowName);
@@ -421,6 +441,7 @@ vtkOpenGLRenderWindow::vtkOpenGLRenderWindow()
   // this->DepthRenderBufferObject = 0;
   this->AlphaBitPlanes = 8;
   this->Capabilities = nullptr;
+  this->RenderBufferTargetDepthSize = 32;
 
   this->TQuad2DVBO = nullptr;
   this->NoiseTextureObject = nullptr;
@@ -480,6 +501,75 @@ vtkOpenGLRenderWindow::~vtkOpenGLRenderWindow()
   this->State->Delete();
 }
 
+#if !(defined(__APPLE__) || defined(__ANDROID__) || defined(__EMSCRIPTEN__))
+//------------------------------------------------------------------------------
+vtkOpenGLRenderWindow* vtkOpenGLRenderWindow::New()
+{
+  const char* backend = std::getenv("VTK_DEFAULT_OPENGL_WINDOW");
+#if defined(_WIN32)
+  if ((backend == nullptr) || (std::string(backend) == "vtkWin32OpenGLRenderWindow"))
+  {
+    vtkNew<vtkWin32OpenGLRenderWindow> win32RenderWindow;
+    win32RenderWindow->SetOffScreenRendering(true);
+    win32RenderWindow->Initialize();
+    if (win32RenderWindow->Initialized)
+    {
+      return win32RenderWindow->NewInstance();
+    }
+  }
+#endif
+#if defined(VTK_USE_X)
+  if ((backend == nullptr) || (std::string(backend) == "vtkXOpenGLRenderWindow"))
+  {
+    gladLoaderLoadGLX(nullptr, 0); // Load core glx functions.
+    // No need to complain if GLX failed to load because vtkXOpenGLRenderWindow will
+    // print the exact reason as a warning anyway.
+    vtkNew<vtkXOpenGLRenderWindow> xRenderWindow;
+    xRenderWindow->SetOffScreenRendering(true);
+    xRenderWindow->Initialize();
+    if (xRenderWindow->Initialized)
+    {
+      return xRenderWindow->NewInstance();
+    }
+  }
+#endif
+#if defined(VTK_OPENGL_HAS_EGL)
+  if ((backend == nullptr) || (std::string(backend) == "vtkEGLRenderWindow"))
+  {
+    // Load core egl functions.
+    if (!gladLoaderLoadEGL(EGL_NO_DISPLAY))
+    {
+      vtkGenericWarningMacro(<< "Failed to load EGL! Please install the EGL library from your "
+                                "distribution's package manager.");
+    }
+    else
+    {
+      vtkNew<vtkEGLRenderWindow> eglRenderWindow;
+      eglRenderWindow->Initialize();
+      if (eglRenderWindow->Initialized)
+      {
+        return eglRenderWindow->NewInstance();
+      }
+    }
+  }
+#endif
+  if ((backend == nullptr) || (std::string(backend) == "vtkOSOpenGLRenderWindow"))
+  {
+    // OSMesa support is always built, don't check for initialization it might work if user has
+    // libOSMesa.so or osmesa.dll.
+    return vtkOSOpenGLRenderWindow::New();
+  }
+  if (backend != nullptr)
+  {
+    vtkGenericWarningMacro(<< "Failed to create a vtkOpenGLRenderWindow subclass with "
+                              "VTK_DEFAULT_OPENGL_WINDOW="
+                           << backend);
+  }
+  // OSMesa support is always built, it might work if user has libOSMesa.so or osmesa.dll.
+  return vtkOSOpenGLRenderWindow::New();
+}
+#endif
+
 //------------------------------------------------------------------------------
 const char* vtkOpenGLRenderWindow::ReportCapabilities()
 {
@@ -692,34 +782,44 @@ void vtkOpenGLRenderWindow::OpenGLInitContext()
   // When a new OpenGL context is created, force an update
   if (!this->Initialized)
   {
-#ifdef GLEW_OK
-    GLenum result = glewInit();
-    this->GlewInitValid = (result == GLEW_OK);
-    if (!this->GlewInitValid)
+#if defined(GLAD_GL)
+    if (this->SymbolLoader.LoadFunction != nullptr)
     {
-      const char* errorMsg = reinterpret_cast<const char*>(glewGetErrorString(result));
-      vtkErrorMacro("GLEW could not be initialized: " << errorMsg);
-      return;
+      if (gladLoadGLUserPtr(this->SymbolLoader.LoadFunction, this->SymbolLoader.UserData) > 0)
+      {
+        this->Initialized = true;
+      }
+      else
+      {
+        vtkWarningMacro(<< "Failed to initialize OpenGL functions!");
+      }
     }
-
-    if (!GLEW_VERSION_3_2 && !GLEW_VERSION_3_1)
+    else
     {
-      vtkErrorMacro("Unable to find a valid OpenGL 3.2 or later implementation. "
-                    "Please update your video card driver to the latest version. "
-                    "If you are using Mesa please make sure you have version 11.2 or "
-                    "later and make sure your driver in Mesa supports OpenGL 3.2 such "
-                    "as llvmpipe or openswr. If you are on windows and using Microsoft "
-                    "remote desktop note that it only supports OpenGL 3.2 with nvidia "
-                    "quadro cards. You can use other remoting software such as nomachine "
-                    "to avoid this issue.");
-      return;
+      if (gladLoaderLoadGL() > 0)
+      {
+        this->Initialized = true;
+      }
+      else
+      {
+        vtkWarningMacro(<< "Failed to initialize OpenGL functions!");
+      }
     }
-#else
-    // GLEW is not being used, so avoid false failure on GL checks later.
-    this->GlewInitValid = true;
-#endif
+#else // gles
     this->Initialized = true;
-
+#endif
+    if (!this->Initialized)
+    {
+      vtkWarningMacro(<< "Unable to find a valid OpenGL 3.2 or later implementation. "
+                         "Please update your video card driver to the latest version. "
+                         "If you are using Mesa please make sure you have version 11.2 or "
+                         "later and make sure your driver in Mesa supports OpenGL 3.2 such "
+                         "as llvmpipe or openswr. If you are on windows and using Microsoft "
+                         "remote desktop note that it only supports OpenGL 3.2 with nvidia "
+                         "quadro cards. You can use other remoting software such as nomachine "
+                         "to avoid this issue.");
+      return;
+    }
     // get this system's supported maximum line width
     // we do it here and store it to avoid repeated glGet
     // calls when the result should not change
@@ -862,6 +962,15 @@ int vtkOpenGLRenderWindow::GetColorBufferSizes(int* rgba)
     {
       attachment = GL_BACK_LEFT;
     }
+    if (attachment == GL_NONE)
+    {
+      // when using vtkGenericOpenGLRenderWindow through QVTKOpenGLNativeWidget,
+      // or a subclass of QOpenGLWidget, the rendering takes place in an offscreen buffer
+      // and is then transferred to the default framebuffer object which is setup
+      // with glDrawBuffers(GL_NONE). So treat it as if it were GL_BACK_LEFT
+      // before querying the color buffer sizes.
+      attachment = GL_BACK_LEFT;
+    }
 
     // make sure we clear any errors before we start
     // otherwise we may get incorrect results
@@ -928,7 +1037,7 @@ int vtkOpenGLRenderWindow::GetColorBufferInternalFormat(int attachmentPoint)
   int format = 0;
 
 #ifndef GL_ES_VERSION_3_0
-  if (GLEW_ARB_direct_state_access)
+  if (GLAD_GL_ARB_direct_state_access)
   {
     int type;
     glGetFramebufferAttachmentParameteriv(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0 + attachmentPoint,
@@ -1232,6 +1341,13 @@ void vtkOpenGLRenderWindow::End()
   this->GetState()->PopFramebufferBindings();
 }
 
+//------------------------------------------------------------------------------
+void vtkOpenGLRenderWindow::SetOpenGLSymbolLoader(VTKOpenGLLoaderFunction loader, void* userData)
+{
+  this->SymbolLoader.LoadFunction = loader;
+  this->SymbolLoader.UserData = userData;
+}
+
 void vtkOpenGLRenderWindow::TextureDepthBlit(vtkTextureObject* source, int srcX, int srcY,
   int srcX2, int srcY2, int destX, int destY, int destX2, int destY2)
 {
@@ -1347,8 +1463,15 @@ void vtkOpenGLRenderWindow::Frame()
     this->RenderFramebuffer->Bind(GL_READ_FRAMEBUFFER);
     this->RenderFramebuffer->ActivateReadBuffer(0);
 
-    this->GetState()->vtkglBlitFramebuffer(0, 0, fbsize[0], fbsize[1], 0, 0, fbsize[0], fbsize[1],
-      (copiedColor ? 0 : GL_COLOR_BUFFER_BIT) | GL_DEPTH_BUFFER_BIT, GL_NEAREST);
+    if (this->FramebufferFlipY)
+    {
+      this->TextureDepthBlit(this->RenderFramebuffer->GetDepthAttachmentAsTextureObject());
+    }
+    else
+    {
+      this->GetState()->vtkglBlitFramebuffer(0, 0, fbsize[0], fbsize[1], 0, 0, fbsize[0], fbsize[1],
+        (copiedColor ? 0 : GL_COLOR_BUFFER_BIT) | GL_DEPTH_BUFFER_BIT, GL_NEAREST);
+    }
 
     this->GetState()->vtkglViewport(0, 0, this->Size[0], this->Size[1]);
     this->GetState()->vtkglScissor(0, 0, this->Size[0], this->Size[1]);
@@ -1364,6 +1487,10 @@ void vtkOpenGLRenderWindow::Frame()
       {
         this->BlitDisplayFramebuffer();
       }
+      if (this->FrameBlitMode == BlitToCurrentWithDepth)
+      {
+        this->BlitDisplayFramebufferColorAndDepth();
+      }
     }
   }
 }
@@ -1549,6 +1676,12 @@ void vtkOpenGLRenderWindow::BlitDisplayFramebuffer()
     this->Size[1], GL_COLOR_BUFFER_BIT, GL_NEAREST);
 }
 
+void vtkOpenGLRenderWindow::BlitDisplayFramebufferColorAndDepth()
+{
+  this->BlitDisplayFramebuffer(0, 0, 0, this->Size[0], this->Size[1], 0, 0, this->Size[0],
+    this->Size[1], GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT, GL_NEAREST);
+}
+
 void vtkOpenGLRenderWindow::BlitDisplayFramebuffer(int right, int srcX, int srcY, int srcWidth,
   int srcHeight, int destX, int destY, int destWidth, int destHeight, int bufferMode,
   int interpolation)
@@ -2479,7 +2612,7 @@ int vtkOpenGLRenderWindow::CreateFramebuffers(int width, int height)
 
   if (!this->RenderFramebuffer->GetFBOIndex())
   {
-    // verify that our multisample setting doe snot exceed the hardware
+    // verify that our multisample setting does not exceed the hardware
     if (this->MultiSamples)
     {
 #ifdef GL_MAX_SAMPLES
@@ -2504,8 +2637,8 @@ int vtkOpenGLRenderWindow::CreateFramebuffers(int width, int height)
 #else
       this->MultiSamples ? false : true, // textures
 #endif
-      1, VTK_UNSIGNED_CHAR, // 1 color buffer uchar
-      true, 32,             // depth buffer
+      1, VTK_UNSIGNED_CHAR,                    // 1 color buffer uchar
+      true, this->RenderBufferTargetDepthSize, // depth buffer
       this->MultiSamples, this->StencilCapable != 0);
     this->LastMultiSamples = this->MultiSamples;
     this->GetState()->PopFramebufferBindings();
@@ -2519,9 +2652,9 @@ int vtkOpenGLRenderWindow::CreateFramebuffers(int width, int height)
   {
     this->GetState()->PushFramebufferBindings();
     this->DisplayFramebuffer->PopulateFramebuffer(width, height,
-      true,                 // textures
-      2, VTK_UNSIGNED_CHAR, // 1 color buffer uchar
-      true, 32,             // depth buffer
+      true,                                    // textures
+      2, VTK_UNSIGNED_CHAR,                    // 1 color buffer uchar
+      true, this->RenderBufferTargetDepthSize, // depth buffer
       0, this->StencilCapable != 0);
     this->GetState()->PopFramebufferBindings();
   }
@@ -2534,9 +2667,9 @@ int vtkOpenGLRenderWindow::CreateFramebuffers(int width, int height)
   {
     this->GetState()->PushFramebufferBindings();
     this->ResolveFramebuffer->PopulateFramebuffer(width, height,
-      true,                 // textures
-      1, VTK_UNSIGNED_CHAR, // 1 color buffer uchar
-      true, 32,             // depth buffer
+      true,                                    // textures
+      1, VTK_UNSIGNED_CHAR,                    // 1 color buffer uchar
+      true, this->RenderBufferTargetDepthSize, // depth buffer
       0, this->StencilCapable != 0);
     this->GetState()->PopFramebufferBindings();
   }
@@ -2621,36 +2754,48 @@ int vtkOpenGLRenderWindow::SupportsOpenGL()
   rw->SetDisplayId(this->GetGenericDisplayId());
   rw->SetOffScreenRendering(1);
   rw->Initialize();
-  if (!rw->GlewInitValid)
+  if (!rw->Initialized)
   {
-    this->OpenGLSupportMessage = "glewInit failed for this window, OpenGL not supported.";
+    this->OpenGLSupportMessage =
+      "Failed to initialize OpenGL for this window, OpenGL not supported.";
     rw->Delete();
     vtkOutputWindow::SetInstance(oldOW);
     oldOW->Delete();
     return 0;
   }
 
-#ifdef GLEW_OK
-
-  else if (GLEW_VERSION_3_2 || GLEW_VERSION_3_1)
+#if defined(GLAD_GL)
+  else if (GLAD_GL_VERSION_3_2 || GLAD_GL_VERSION_3_1)
   {
     this->OpenGLSupportResult = 1;
     this->OpenGLSupportMessage = "The system appears to support OpenGL 3.2/3.1";
   }
-#else
-#ifdef GL_ES_VERSION_2_0
-  this->OpenGLSupportResult = 1;
-  this->OpenGLSupportMessage = "The system appears to support OpenGL ES 2.0";
-#endif
-#ifdef GL_ES_VERSION_3_0
-  this->OpenGLSupportResult = 1;
-  this->OpenGLSupportMessage = "The system appears to support OpenGL ES 3.0";
-#endif
+#elif defined(GLAD_GLES2)
+  else if (GLAD_GL_ES_VERSION_3_2)
+  {
+    this->OpenGLSupportResult = 1;
+    this->OpenGLSupportMessage = "The system appears to support OpenGL ES 3.2";
+  }
+  else if (GLAD_GL_ES_VERSION_3_1)
+  {
+    this->OpenGLSupportResult = 1;
+    this->OpenGLSupportMessage = "The system appears to support OpenGL ES 3.1";
+  }
+  else if (GLAD_GL_ES_VERSION_3_0)
+  {
+    this->OpenGLSupportResult = 1;
+    this->OpenGLSupportMessage = "The system appears to support OpenGL ES 3.0";
+  }
+  else if (GLAD_GL_ES_VERSION_2_0)
+  {
+    this->OpenGLSupportResult = 1;
+    this->OpenGLSupportMessage = "The system appears to support OpenGL ES 2.0";
+  }
 #endif
 
   if (this->OpenGLSupportResult)
   {
-    // even if glew thinks we have support we should actually try linking a
+    // even if glad thinks we have support we should actually try linking a
     // shader program to make sure
     vtkShaderProgram* newShader = rw->GetShaderCache()->ReadyShaderProgram(
       // simple vert shader
